#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Some simple tests for the autopaths package.

You can run this file like this:

  ipython -i -- ~/repos/autopaths/test/test_symlink.py

"""

# Built-in modules #
import os, inspect

# Get current directory (works always) #
file_name = os.path.abspath((inspect.stack()[0])[1])
this_dir  = os.path.dirname(os.path.abspath(file_name)) + '/'

# All our example file system #
dummy_files = this_dir + 'dummy_file_system/'

# Internal modules #
from autopaths.dir_path import DirectoryPath

###############################################################################
def test_list_files():
    d = DirectoryPath(dummy_files)
    print(list((f.name, str(f.size)) for f in d.files))

def test_symlink():
    d = DirectoryPath(dummy_files)
    source = d['lorem']
    destin = d + 'ipsum/'
    print(source, destin)
    source.link_to(destin)

###############################################################################
if __name__ == '__main__':
    test_list_files()
    test_symlink()


# Built-in modules #
import os, glob

# Internal modules #
import autopaths

# Constants #
if os.name == "posix": sep = "/"
if os.name == "nt":    sep = "\\"

################################################################################
class BasePath(str):
    """
    This object contains methods that are common to both FilePath objects
    and DirectoryPath objects.
    """

    def __repr__(self): return '<%s object "%s">' % (self.__class__.__name__, self.path)

    @classmethod
    def clean_path(cls, path):
        """Given a path, return a cleaned up version for initialization."""
        # Conserve 'None' object style #
        if path is None: return None
        # Don't nest BasePaths object or the like #
        if hasattr(path, 'path'): path = path.path
        # Expand the tilda #
        if "~" in path: path = os.path.expanduser(path)
        # We will store the path with the OS specific separator #
        # We will never mix both kinds of separators #
        if os.name == "posix": path = path.replace("\\", sep)
        if os.name == "nt":    path = path.replace("/",  sep)
        # Expand star #
        if "*" in path:
            matches = glob.glob(path)
            if len(matches) < 1: raise Exception("Found exactly no paths matching '%s'" % path)
            if len(matches) > 1: raise Exception("Found several paths matching '%s'" % path)
            path = matches[0]
        # Our standard is to end with a slash for directories #
        if cls is autopaths.dir_path.DirectoryPath:
            if not path.endswith(sep):
                path += sep
        # Return the result #
        return path

    def __new__(cls, path, *args, **kwargs):
        """A Path object is in fact a string."""
        return str.__new__(cls, cls.clean_path(path))

    def __init__(self, path):
        self.path = self.clean_path(path)

    def __add__(self, other):
        if os.name == "posix": other = other.replace("\\", sep)
        if os.name == "nt":    other = other.replace("/",  sep)
        if other.endswith(sep): return autopaths.dir_path.DirectoryPath(self.path + other)
        else:                   return autopaths.file_path.FilePath(self.path + other)

    # ------------------------------ Properties ----------------------------- #
    @property
    def short_prefix(self):
        """Just the filename without any extension or periods."""
        return self.name.split('.')[0]

    @property
    def escaped(self):
        """
        The path with special characters escaped.
        For instance a backslash becomes a double backslash.
        """
        return self.path.replace("\\", "\\\\")

    @property
    def unix_style(self):
        """The path with forward slashes and no disk drive."""
        if self.path[1] == ':': path = self.path[2:]
        else:                   path = self.path
        return path.replace("\\", "/")

    @property
    def wsl_style(self):
        """
        The path with forward slashes and a windows subsytem
        for linux style leading disk drive.
        """
        return "/mnt/c" + self.unix_style

    @property
    def win_style(self):
        """The path with backward slashes."""
        return self.path.replace("/", "\\")

    @property
    def exists(self):
        """
        Does it exist in the file system?
        Returns True even for broken symbolic links.
        """
        return os.path.lexists(self.path)

    @property
    def is_symlink(self):
        """Is this file a symbolic link to an other file?"""
        if os.name == "posix": return os.path.islink(self.path)
        if os.name == "nt":
            import win32api
            import win32con
            num = win32con.FILE_ATTRIBUTE_REPARSE_POINT
            return bool(win32api.GetFileAttributes(self.path) & num)

    @property
    def permissions(self):
        """Convenience object for dealing with permissions."""
        return autopaths.file_permissions.FilePermissions(self.path)

    @property
    def mdate(self):
        """Return the modification date as a unix time."""
        return os.path.getmtime(self.path)

    @property
    def mdate_iso(self):
        """Return the modification date as a datetime iso object."""
        return datetime.fromtimestamp(self.mdate).isoformat()

    @property
    def cdate(self):
        """Return the creation date."""
        return os.path.getctime(self.path)

    @property
    def cdate_iso(self):
        """Return the creation date as a datetime iso object."""
        return datetime.fromtimestamp(self.cdate).isoformat()

    #-------------------------------- Methods --------------------------------#
    def link_from(self, path, safe=False, absolute=False):
        """
        Make a link here pointing to another file/directory somewhere else.
        The destination is hence self.path and the source is *path*.
        """
        # Get source and destination #
        from autopaths import Path
        source      = Path(path)
        destination = self.path
        # Call method #
        self._symlink(source, destination, safe, absolute)

    def link_to(self, path, safe=False, absolute=False):
        """
        Create a link somewhere else pointing to this file.
        The destination is hence *path* and the source is self.path.
        """
        # Get source and destination #
        from autopaths import Path
        source      = self.path
        destination = Path(path)
        # Call method #
        self._symlink(source, destination, safe, absolute)

    def _symlink(self, source, destination, safe, absolute):
        # If source is a file and the destination is a dir, put it inside #
        if os.path.isdir(destination) and not os.path.isdir(source):
            destination = destination + source.filename
        # Do we want absolute paths #
        if absolute: source = source.absolute_path
        # Strip trailing separators #
        source      = source.rstrip(sep)
        destination = destination.rstrip(sep)
        # Windows doesn't have os.symlink #
        if os.name == "posix": self.symlinks_on_linux(source, destination, safe)
        if os.name == "nt":    self.symlinks_on_windows(source, destination, safe)

    def symlinks_on_linux(self, source, destination, safe):
        # Do it unsafely #
        if not safe:
            if os.path.exists(destination): os.remove(destination)
            os.symlink(source, destination)
        # Do it safely #
        if safe:
            try: os.remove(destination)
            except OSError: pass
            try: os.symlink(source, destination)
            except OSError: pass

    def symlinks_on_windows(self, source, destination, safe):
        """Yes, source and destination need to be in the reverse order"""
        import win32file
        if os.path.isdir(source):
            return win32file.CreateSymbolicLink(destination, source, 1)
        else:
            return win32file.CreateSymbolicLink(destination, source, 0)



# Built-in modules #
import os, tempfile, subprocess, shutil, codecs, gzip, zipfile, hashlib

# Internal modules #
import autopaths
from autopaths.common import pad_extra_whitespace

# Constants #
if os.name == "posix": sep = "/"
if os.name == "nt":    sep = "\\"

# Python 2 compatibility #
import six
if six.PY2:
    from io import open

################################################################################
class FilePath(autopaths.base_path.BasePath):
    """
    Holds a string pointing to a file path and adds methods to interact with
    files on disk.

    I can never remember all those darn `os.path` commands, so I made a class
    that wraps them with an easier and more pythonic syntax.

        path = FilePath('/home/root/text.txt')
        print path.extension
        print path.directory
        print path.filename

    You can find lots of the common things you would need to do with file paths.
    Such as: path.make_executable() etc etc.
    """

    def __bool__(self): return self.path is not None and self.count_bytes != 0
    __nonzero__ = __bool__

    def __iter__(self):
        with open(self.path, 'r', encoding='utf-8') as handle:
            for line in handle: yield line

    def __len__(self):
        if self.path is None: return 0
        return self.count

    def __sub__(self, directory):
        """
        Subtract a directory from the current path to get the relative path
        of the current file from that directory.
        """
        return os.path.relpath(self.path, directory)

    def __enter__(self):
        """Called when entering the 'with' statement (context manager)."""
        return self

    def __exit__(self, errtype, value, traceback):
        """
        Called when exiting the 'with' statement.
        This enables us to close the file or database properly, even when
        exceptions are raised.
        """
        self.close()

    # ------------------------------ Properties ----------------------------- #
    @property
    def first(self):
        """Just the first line. Don't try this on binary files."""
        with open(self.path, 'r') as handle:
            for line in handle: return line

    @property
    def prefix_path(self):
        """The full path without the (last) extension and trailing period."""
        return str(os.path.splitext(self.path)[0])

    @property
    def prefix(self):
        """Just the filename without the (last) extension and trailing period."""
        return str(os.path.basename(self.prefix_path))

    @property
    def name(self):
        """Shortcut for self.filename."""
        return self.filename

    @property
    def filename(self):
        """Just the filename with the extension."""
        return str(os.path.basename(self.path))

    @property
    def extension(self):
        """The extension with the leading period."""
        return os.path.splitext(self.path)[1]

    @property
    def directory(self):
        """The directory containing this file."""
        # The built-in function #
        directory = os.path.dirname(self.path)
        # Maybe we need to go the absolute path way #
        if not directory: directory = os.path.dirname(self.absolute_path)
        # Return #
        return autopaths.dir_path.DirectoryPath(directory)

    @property
    def count_bytes(self):
        """The number of bytes."""
        if not self.exists: return 0
        return os.path.getsize(self.path)

    @property
    def count(self):
        """We are going to default to the number of lines."""
        # Third party modules #
        if os.name == "posix": import sh
        if os.name == "nt":    import pbs
        # Count lines #
        if os.name == "posix": return int(sh.wc('-l', self.path).split()[0])
        if os.name == "nt":    return int(pbs.Command("find")('/c', '/v', '""', self.path))

    @property
    def size(self):
        """Human readable file size."""
        return autopaths.file_size.FileSize(self.count_bytes)

    @property
    def contents(self):
        """The contents as a string."""
        with open(self.path, 'r') as handle: return handle.read()

    @property
    def contents_utf8(self):
        """The contents as a unicode string."""
        with open(self.path, encoding='utf8') as handle:return handle.read()

    @property
    def absolute_path(self):
        """The absolute path starting with a `/`."""
        return FilePath(os.path.abspath(self.path))

    @property
    def physical_path(self):
        """The physical path like in `pwd -P`."""
        return FilePath(os.path.realpath(self.path))

    @property
    def relative_path(self):
        """The relative path when compared with current directory."""
        return FilePath(os.path.relpath(self.physical_path))

    def rel_path_from(self, path):
        """The relative path when compared to the given path."""
        return FilePath(os.path.relpath(self.path, path))

    @property
    def md5(self):
        """Compute the md5 of a file. Pretty fast."""
        md5 = hashlib.md5()
        with open(self.path, "rb") as f:
            for block in iter(lambda: f.read(4096), b""):
                md5.update(block)
        return md5.hexdigest()

    @property
    def might_be_binary(self):
        """Try to quickly guess if the file is binary."""
        from binaryornot.check import is_binary
        return is_binary(self.path)

    @property
    def contains_binary(self):
        """Return True if the file contains binary characters."""
        from binaryornot.helpers import is_binary_string
        return is_binary_string(self.contents)

    @property
    def magic_number(self):
        """Return the first four bytes of the file."""
        with open(self.path, 'rb') as f: return f.read(4)

    @property
    def lines(self):
        """Get all lines in a list with \n striped."""
        with open(self.path, 'r', encoding='utf-8') as handle:
            return [line.strip('\n') for line in handle]

    #-------------------------------- Methods --------------------------------#
    def read(self, encoding=None):
        with open(self.path, 'r', encoding) as handle: content = handle.read()
        return content

    def create(self):
        if not self.directory.exists: self.directory.create()
        self.open('w')
        return self

    def touch(self):
        """Just create an empty file if it does not exist."""
        with open(self.path, 'a'): os.utime(self.path, None)

    def open(self, mode='r'):
        self.handle = open(self.path, mode)
        return self.handle

    def add_str(self, string):
        self.handle.write(string)

    def close(self):
        self.handle.close()

    def write(self, content, encoding=None, mode='w'):
        if encoding is None:
            with open(self.path, mode) as handle: handle.write(content)
        else:
            with open(self.path, mode, encoding=encoding) as handle: handle.write(content)

    def writelines(self, content, encoding=None, mode='w'):
        if encoding is None:
            with open(self.path, mode) as handle: handle.writelines(content)
        else:
            with open(self.path, mode, encoding=encoding) as handle: handle.writelines(content)

    def remove(self):
        if not self.exists: return False
        os.remove(self.path)
        return True

    def copy(self, path):
        """Copy to a different path."""
        # Directory special case #
        if path.endswith(sep): path += self.filename
        # Normal case #
        shutil.copy2(self.path, path)

    def execute(self):
        return subprocess.call([self.path])

    def replace_extension(self, new_extension='txt'):
        """Return a new path with the extension swapped out."""
        return FilePath(os.path.splitext(self.path)[0] + '.' + new_extension)

    def new_name_insert(self, string):
        """Return a new name by appending a string before the extension."""
        return self.prefix_path + "." + string + self.extension

    def make_directory(self):
        """Create the directory the file is supposed to be in if it does not exist."""
        if not self.directory.exists: self.directory.create()

    def must_exist(self):
        """Raise an exception if the path doesn't exist."""
        if not self.exists: raise Exception("The file path '%s' does not exist." % self.path)

    def head(self, num_lines=20):
        """Yield the first few lines."""
        lines = iter(self)
        for _ in range(num_lines): yield next(lines)

    @property
    def pretty_head(self):
        return "\n" + pad_extra_whitespace("\n".join(self.head()), 4) + "\n"

    def tail(self, num_lines=20, encoding='utf-8'):
        """Yield the last few lines of the file."""
        # Constant #
        buffer_size = 1024
        # Smart algorithm #
        with open(self.path, 'rb') as handle:
            handle.seek(0, 2)
            num_bytes = handle.tell()
            size      = num_lines + 1
            block     = -1
            data      = []
            # Loop #
            while size > 0 and num_bytes > 0:
                if num_bytes - buffer_size > 0:
                    # Seek back one whole buffer_size #
                    handle.seek(block * buffer_size, os.SEEK_END)
                    # Read buffer #
                    data.insert(0, handle.read(buffer_size).decode(encoding))
                else:
                    # File too small, start from beginning #
                    handle.seek(0,0)
                    # Only read what was not read #
                    data.insert(0, handle.read(num_bytes).decode(encoding))
                lines_found = data[0].count('\n')
                size       -= lines_found
                num_bytes  -= buffer_size
                block      -= 1
            # Return #
            for line in ''.join(data).splitlines()[-num_lines:]: yield line

    @property
    def pretty_tail(self):
        return "\n" + pad_extra_whitespace("\n".join(self.tail()), 4) + "\n"

    def move_to(self, path, overwrite=False):
        """Move the file to a new location."""
        # Special directory case, keep the same name (put it inside) #
        if path.endswith(sep): path = path + self.filename
        # Normal case #
        if os.path.exists(path) and overwrite: os.remove(path)
        assert not os.path.exists(path)
        shutil.move(self.path, path)
        # Update the internal link #
        self.path = path

    def rename(self, new_name):
        """Rename the file but leave it in the same directory."""
        assert sep not in new_name
        path = self.directory + new_name
        assert not os.path.exists(path)
        shutil.move(self.path, path)
        # Update the internal link #
        self.path = path

    #------------------------------ Compression ------------------------------#
    def gzip_to(self, path=None):
        """Make a gzipped version of the file at a given path."""
        if path is None: path = self.path + ".gz"
        with open(self.path, 'rb') as orig_file:
            with gzip.open(path, 'wb') as new_file:
                new_file.writelines(orig_file)
        return FilePath(path)

    def ungzip_to(self, path=None, mode='w'):
        """Make an unzipped version of the file at a given path."""
        if path is None: path = self.path[:3]
        with gzip.open(self, 'rb') as orig_file:
            with open(path, mode) as new_file:
                new_file.writelines(orig_file)
        return FilePath(path)

    def zip_to(self, path=None):
        """Make a zipped version of the file at a given path."""
        pass

    def unzip_to(self, destination=None, inplace=False, single=True):
        """
        Unzip a standard zip file. Can specify the destination of the
        uncompressed file, or just set inplace=True to delete the original.
        """
        # Check #
        assert zipfile.is_zipfile(self.path)
        # Load #
        z = zipfile.ZipFile(self.path)
        if single or inplace: assert len(z.infolist()) == 1
        # Single file #
        if single:
            member = z.infolist()[0]
            tmpdir = tempfile.mkdtemp() + sep
            z.extract(member, tmpdir)
            z.close()
            if inplace: shutil.move(tmpdir + member.filename, self.directory + member.filename)
            else:       shutil.move(tmpdir + member.filename, destination)
        # Multifile - no security, dangerous - Will use CWD if dest is None!! #
        # If a file starts with an absolute path, will overwrite your files anywhere #
        if not single:
            z.extractall(destination)

    def targz_to(self, path=None):
        """Make a targzipped version of the file at a given path."""
        pass

    def untargz_to(self, destination=None):
        """Make an untargzipped version of the file at a given path."""
        import tarfile
        archive = tarfile.open(self.path, 'r:gz')
        archive.extractall(destination)

    def tar_to(self, path=None):
        """Make a tared version of the file at a given path."""
        pass

    def untar_to(self, destination=None):
        """Make an untared version of the file at a given path."""
        import tarfile
        with tarfile.open(self.path) as archive: archive.extractall(destination)

    #-------------------------------- Modify ---------------------------------#
    def append(self, data):
        """Append some text or an other file to the current file."""
        if isinstance(data, FilePath): data = data.contents
        with open(self.path, "a") as handle: handle.write(data)

    def prepend(self, data, buffer_size=1 << 15):
        """
        Prepend some text or an other file to the current file.
        TODO:
        * Add a random string to the backup file.
        * Restore permissions after copy.
        """
        # Check there is something to prepend #
        assert data
        # Support passing other files #
        if isinstance(data, FilePath): data = data.contents
        # Create a new file #
        result_file = autopaths.tmp_path.new_temp_file()
        # Open input/output files, note: output file's permissions lost #
        with open(self) as in_handle:
            with open(result_file, 'w') as out_handle:
                while data:
                    out_handle.write(data)
                    data = in_handle.read(buffer_size)
        # Switch the files around #
        self.remove()
        result_file.move_to(self)

    def remove_line(self, line_to_remove):
        """Search the file for a given line, and if found, remove it."""
        # Check there is something to remove #
        assert line_to_remove
        # Create a new file #
        result_file = autopaths.tmp_path.new_temp_file()
        # Open input/output files, note: output file's permissions lost #
        result_file.writelines(line for line in self if line != line_to_remove)
        # Switch the files around #
        self.remove()
        result_file.move_to(self)

    def remove_first_line(self):
        """
        Remove the first line of the file.
        Equivalent to sh.sed('-i', '1d', self.path)
        """
        # Create a new file #
        result_file = autopaths.tmp_path.new_temp_file()
        # Open input/output files, note: output file's permissions lost #
        all_lines = iter(self)
        next(all_lines)
        result_file.writelines(all_lines)
        # Switch the files around #
        self.remove()
        result_file.move_to(self)

    def replace_line(self, line_to_remove, line_to_insert):
        """
        Search the file for a given line, and if found,
        replace it with another line.
        """
        # Check the line endings #
        line_to_remove = line_to_remove.strip('\n')
        line_to_insert = line_to_insert.strip('\n')
        # Create a new file #
        result_file = autopaths.tmp_path.new_temp_file()
        # Generate the lines #
        def new_lines():
            for line in self:
                if line.strip() == line_to_remove: yield line_to_insert + '\n'
                else:                              yield line
        # Open input/output files, note: output file's permissions lost #
        result_file.writelines(new_lines())
        # Switch the files around #
        self.remove()
        result_file.move_to(self)

    def replace_word(self, word_to_find, replacement_word):
        """
        Search the file for a given word, and if found,
        replace every occurrence of it with another word.
        """
        # Create a new file #
        result_file = autopaths.tmp_path.new_temp_file()
        # Generate the lines #
        def new_lines():
            for line in self:
                yield line.replace(word_to_find, replacement_word)
        # Open input/output files, note: output file's permissions lost #
        result_file.writelines(new_lines())
        # Switch the files around #
        self.remove()
        result_file.move_to(self)

    #---------------------------- External tools -----------------------------#
    def sed_replace(self, before, after):
        if os.name == "posix":
            import sh
            return sh.sed('-i', 's/%s/%s/' % (before, after), self.path)
        if os.name == "nt":
            import pbs
            sed_cmd = 'sed -i "s/%s/%s/" %s' % (before, after, self.path)
            return pbs.bash('-c', "'" + sed_cmd + "'" )



# Built-in modules #
import os, shutil, glob

# Internal modules #
import autopaths

# Constants #
if os.name == "posix": sep = "/"
if os.name == "nt":    sep = "\\"

################################################################################
class DirectoryPath(autopaths.base_path.BasePath):
    """
    Holds a string pointing to a directory path and adds methods to interact with
    directories.
    """

    def __len__(self): return len(list(self.flat_contents))

    def __iter__(self): return self.flat_contents

    def __contains__(self, item): return item in [x.name for x in self.flat_contents]

    def __getitem__(self, item):
        for path in self.flat_contents:
            if path.name == item:
                return path
        raise KeyError("Couldn't find '%s' in '%s'" % (item, self.path))

    # ------------------------------ Properties ----------------------------- #
    @property
    def p(self):
        if not hasattr(self, 'all_paths'):
            raise Exception("You need to define 'all_paths' to use this function")
        return autopaths.auto_paths.AutoPaths(self.path, self.all_paths)

    @property
    def name(self):
        """Just the directory name."""
        return os.path.basename(os.path.dirname(self.path))

    @property
    def prefix_path(self):
        """The full path without the extension."""
        return os.path.splitext(self.path)[0].rstrip(sep)

    @property
    def absolute_path(self):
        """The absolute path starting with a `/`."""
        return os.path.abspath(self.path) + sep

    @property
    def directory(self):
        """The full path of the directory containing this one (parent dir)."""
        # The built-in function #
        directory = os.path.dirname(os.path.dirname(self.path))
        # Maybe we need to go the absolute path way #
        if not directory:
            directory = os.path.dirname(os.path.dirname(self.absolute_path))
        # Return #
        return autopaths.dir_path.DirectoryPath(directory)

    @property
    def empty(self):
        """Does the directory contain no files?"""
        return len(list(self.flat_contents)) == 0

    @property
    def size(self):
        """The total size in bytes of all file contents."""
        return autopaths.file_size.FileSize(sum(f.count_bytes for f in self.files))

    #-------------------------- Recursive contents ---------------------------#
    @property
    def contents(self):
        """The files and directories in this directory, recursively."""
        for root, dirs, files in os.walk(self.path, topdown=False):
            for d in dirs:  yield DirectoryPath(os.path.join(root, d))
            for f in files: yield autopaths.file_path.FilePath(os.path.join(root, f))

    @property
    def files(self):
        """The files in this directory, recursively."""
        for root, dirs, files in os.walk(self.path, topdown=False):
            for f in files: yield autopaths.file_path.FilePath(os.path.join(root, f))

    @property
    def directories(self):
        """The directories in this directory, recursively."""
        for root, dirs, files in os.walk(self.path, topdown=False):
            for d in dirs: yield DirectoryPath(os.path.join(root, d))

    #----------------------------- Flat contents -----------------------------#
    @property
    def flat_contents(self):
        """The files and directories in this directory non-recursively."""
        for root, dirs, files in os.walk(self.path):
            for d in dirs:  yield DirectoryPath(os.path.join(root, d))
            for f in files: yield autopaths.file_path.FilePath(os.path.join(root, f))
            break

    @property
    def flat_files(self):
        """The files in this directory non-recursively, and sorted.
        #TODO: check for permission denied in directory."""
        for root, dirs, files in os.walk(self.path):
            result = [autopaths.file_path.FilePath(os.path.join(root, f)) for f in files]
            break
        else:
            result = []
        result.sort(key=lambda x: autopaths.common.natural_sort(x.path))
        return result

    @property
    def flat_directories(self):
        """The directories in this directory non-recursively, and sorted."""
        for root, dirs, files in os.walk(self.path):
            result = [DirectoryPath(os.path.join(root, d)) for d in dirs]
            break
        else:
            result = []
        result.sort(key=lambda x: autopaths.common.natural_sort(x.path))
        return result

    #------------------------------- Methods ---------------------------------#
    def must_exist(self):
        """Raise an exception if the directory doesn't exist."""
        if not os.path.isdir(self.path):
            raise Exception("The directory path '%s' does not exist." % self.path)

    def remove(self, safe=True):
        if not self.exists: return False
        if self.is_symlink: return self.remove_when_symlink()
        shutil.rmtree(self.path, ignore_errors=safe)
        return True

    def remove_when_symlink(self):
        if not self.exists: return False
        os.remove(self.path.rstrip(sep))
        return True

    def create(self, safe=False, inherit=True):
        # Create it #
        if not safe:
            os.makedirs(self.path)
            if inherit: os.chmod(self.path, self.directory.permissions.number)
        if safe:
            try:
                os.makedirs(self.path)
                if inherit: os.chmod(self.path, self.directory.permissions.number)
            except OSError: pass

    def create_if_not_exists(self):
        if not self.exists: self.create()

    def move_to(self, path):
        """Move the directory."""
        # Check #
        assert not os.path.exists(path)
        shutil.move(self.path, path)
        # Update the internal link #
        self.path = path

    def zip(self, keep_orig=False):
        """Make a zip archive of the directory"""
        shutil.make_archive(self.prefix_path , "zip", self.directory, self.name)
        if not keep_orig: self.remove()

    def copy(self, path):
        assert not os.path.exists(path)
        shutil.copytree(str(self.path), str(path))

    def glob(self, pattern):
        """Perform a glob search in this directory."""
        files = glob.glob(self.path + pattern)
        return map(autopaths.file_path.FilePath, files)

    def find(self, pattern):
        """Find a file in this directory."""
        f = glob.glob(self.path + pattern)[0]
        return autopaths.file_path.FilePath(f)

    def unnest(self):
        """
        Move all contents (files and directories) of this directory
        to its parent directory, and remove this directory.
        """
        for item in self.flat_contents: item.move_to(self.directory)
        self.remove()

    def zip_to(self, path=None):
        """Make a zipped version of the directory at a given path."""
        # Case where path is None #
        if path is None: path = self.directory + self.name + ".zip"
        # The method adds .zip so let's remove it #
        if path.endswith('.zip'): path_no_ext = path.prefix_path
        else:                     path_no_ext = path
        # Compress #
        shutil.make_archive(path_no_ext, 'zip', self.path)
        # Return #
        return path_no_ext


# Built-in modules #
from math import log

################################################################################
class FileSize(object):
    """
    Container for a size in bytes with a human readable representation
    Use it like this:

        >>> from autopaths.file_size import FileSize
        >>> size = FileSize(123123123)
        >>> print size
        '123.1 MB'
        >>> size = FileSize(123123123, system='binary')
        >>> print size
        '117.4 MiB'
    """

    precisions = [0, 0, 1, 2, 2, 2]

    def __init__(self, size, system='decimal'):
        # Record the size #
        self.size = size
        # Pick the system used #
        if system is 'binary':
            self.chunk = 1024
            self.units = ['bytes', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB']
        elif system is 'decimal':
            self.chunk = 1000
            self.units = ['bytes', 'KB', 'MB', 'GB', 'TB', 'PB']
        else:
            raise Exception("Unrecognized file size system '%s'" % system)

    def __eq__(self, other):
        return self.size == other

    def __str__(self):
        # Special cases #
        if self.size == 0: return '0 bytes'
        if self.size == 1: return '1 byte'
        # Pick the right unit #
        unit = self.units[min(int(log(self.size, self.chunk)), len(self.units) - 1)]
        # Compute #
        exponent  = self.units.index(unit)
        quotient  = float(self.size) / self.chunk**exponent
        precision = self.precisions[exponent]
        template  = '{:.%sf} {}' % (precision)
        string    =  template.format(quotient, unit)
        # Return a string #
        return string



# Special variables #
__version__ = '1.3.8'

# Built in modules #
import os

# This is done to avoid circular import errors #
from autopaths import common
from autopaths import file_size
from autopaths import file_permissions
from autopaths import base_path
from autopaths import file_path
from autopaths import dir_path
from autopaths import tmp_path
from autopaths import auto_paths

###############################################################################
def Path(path):
    # Cast to string #
    path = str(path)
    # Return either a file or a directory path #
    if os.path.isdir(path) or path.endswith('/'):
        return dir_path.DirectoryPath(path)
    else:
        return file_path.FilePath(path)



# Built-in modules #
import os, tempfile, re

# Internal modules #
import autopaths

# Constants #
if os.name == "posix": sep = "/"
if os.name == "nt":    sep = "\\"

################################################################################
class AutoPaths(object):
    """
    You can use this class like this when making pipelines:

        class Sample(object):
            all_paths = '''
                /raw/raw.sff
                /raw/raw.fastq
                /clean/trim.fastq
                /clean/clean.fastq'''

            def __init__(self, base_dir):
                self.p = AutoPaths(base_dir, self.all_paths)

            def clean(self):
                shutil.move(self.p.raw_sff, self.p.clean_fastq)
    """

    def __repr__(self): return '<%s object on "%s">' % (self.__class__.__name__, self._base_dir)

    def __init__(self, base_dir, all_paths):
        # Don't nest Path objects or the like #
        if hasattr(base_dir, 'path'): base_dir = base_dir.path
        # Attributes #
        self._base_dir  = base_dir
        self._all_paths = all_paths
        self._tmp_dir   = tempfile.gettempdir() + sep
        # Parse input #
        self._paths = [p for p in all_paths.split('\n') if p.strip(' ')]
        self._paths = [PathItems(p.strip(' '), base_dir) for p in self._paths]

    def __call__(self, key):    return self.__getattr__(key)
    def __getitem__(self, key): return self.__getattr__(key)
    def get(self, key):         return self.__getattr__(key)

    def __getattr__(self, key):
        # Let built-ins pass through to object #
        if key.startswith('__') and key.endswith('__'): return object.__getattr__(key)
        # Special cases that should do to the actual dictionary #
        if key.startswith('_'): return self.__dict__[key]
        # Temporary items #
        if key == 'tmp_dir': return self.__dict__['_tmp_dir']
        if key == 'tmp':     return self.__dict__['tmp']
        # Search #
        items = key.split('_')
        # Is it a directory ? #
        if items[-1] == 'dir':
            items.pop(-1)
            return self.search_for_dir(key, items)
        else:
            return self.search_for_file(key, items)

    def __iter__(self):
        """Yield all the paths we have stored."""
        for path in self._paths: yield autopaths.Path(path.complete_path)

    def search_for_file(self, key, items):
        # Search #
        matches = [set([p for p in self._paths if i in p]) for i in items]
        result = set.intersection(*matches)
        # No matches #
        if len(result) == 0:
            raise Exception("Could not find any path matching '%s'" % key)
        # Multiple matches, advantage file name #
        if len(result) > 1:
            best_score = max([p.score_file(items) for p in result])
            result = [p for p in result if p.score_file(items) >= best_score]
        # Multiple matches, take the one with less parts #
        if len(result) > 1:
            shortest = min([len(p) for p in result])
            result = [p for p in result if len(p) <= shortest]
        # Multiple matches, error #
        if len(result) > 1:
            raise Exception("Found several paths matching '%s'" % key)
        # Make the directory #
        result = result.pop()
        directory = autopaths.dir_path.DirectoryPath(result.complete_dir)
        if not directory.exists: directory.create(safe=True)
        # Return different types #
        from plumbing.csv_tables import CSVTable, TSVTable
        if result.complete_path.endswith('.tsv'): return TSVTable(result.complete_path)
        if result.complete_path.endswith('.csv'): return CSVTable(result.complete_path)
        # Return base case #
        return autopaths.file_path.FilePath(result.complete_path)

    def search_for_dir(self, key, items):
        # Search #
        matches = [set([p for p in self._paths if i in p]) for i in items]
        result = set.intersection(*matches)
        # No matches #
        if len(result) == 0:
            raise Exception("Could not find any path matching '%s'" % key)
        # Multiple matches, advantage dir name #
        if len(result) > 1:
            best_score = max([p.score_dir(items) for p in result])
            result = [p for p in result if p.score_dir(items) >= best_score]
        # Multiple matches, take the one with less parts #
        if len(result) > 1:
            shortest = min([len(p) for p in result])
            result = [p for p in result if len(p) <= shortest]
        # Multiple matches, maybe they all are the same directory #
        if len(result) > 1:
            if len(set([p.dir for p in result])) == 1: result = [result[0]]
        # Multiple matches, error #
        if len(result) > 1:
            raise Exception("Found several paths matching '%s'" % key)
        # Make the directory #
        result = result.pop()
        directory = autopaths.dir_path.DirectoryPath(result.complete_dir)
        if not directory.exists: directory.create(safe=True)
        # Return #
        return directory

    @property
    def tmp_dir(self):
        if not self._tmp_dir: self._tmp_dir = tempfile.mkdtemp() + sep
        return self._tmp_dir

    @property
    def tmp(self):
        return self.tmp_dir + 'autopath.tmp'

################################################################################
class PathItems(object):
    delimiters = '_', '.', '/'
    pattern = re.compile('|'.join(map(re.escape, delimiters)))

    def __repr__(self): return '<%s object "%s">' % (self.__class__.__name__, self.path)

    def __init__(self, path, base_dir):
        self.path           = path
        self.base_dir       = base_dir
        self.dir, self.name = os.path.split(path)
        self.name_items     = self.pattern.split(self.name) if self.name else []
        self.dir_items      = self.pattern.split(self.dir)  if self.dir  else []
        self.all_items      = self.name_items + self.dir_items

    def __contains__(self, i):
        return i in self.all_items

    def __len__(self):
        return len(self.all_items)

    def score_file(self, items):
        return sum([1.0 if i in self.name_items else 0.5 for i in items if i in self])

    def score_dir(self, items):
        return sum([1.0 if i in self.dir_items else 0.5 for i in items if i in self])

    @property
    def complete_path(self):
        return sep + os.path.relpath(self.base_dir + self.path, sep)

    @property
    def complete_dir(self):
        return sep + os.path.relpath(self.base_dir + self.dir, sep) + sep



# Built-in modules #
import os, stat

################################################################################
class FilePermissions(object):
    """Container for reading and setting a files permissions."""

    def __init__(self, path):
        self.path = path

    @property
    def number(self):
        """The permission bits as an octal integer."""
        return os.stat(self.path).st_mode & 0o0777

    @property
    def is_executable(self):
        """Can you execute this file?"""
        return os.access(self.path, os.X_OK)

    def make_executable(self):
        return os.chmod(self.path, os.stat(self.path).st_mode | stat.S_IEXEC)

    def only_readable(self):
        """Remove all writing privileges."""
        return os.chmod(self.path, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)



# Built-in modules #
import tempfile

# Internal modules #
from autopaths.file_path import FilePath
from autopaths.dir_path  import DirectoryPath

################################################################################
def new_temp_handle(**kwargs):
    """A new temporary handle ready to be written to."""
    handle = tempfile.NamedTemporaryFile(delete=False, **kwargs)
    return handle

################################################################################
def new_temp_path(**kwargs):
    """A new temporary path."""
    handle = tempfile.NamedTemporaryFile(**kwargs)
    path   = handle.name
    handle.close()
    return path

################################################################################
def new_temp_file(**kwargs):
    """A new temporary path as a FilePath object."""
    handle = tempfile.NamedTemporaryFile(delete=False, **kwargs)
    path   = handle.name
    handle.close()
    return FilePath(path)

################################################################################
def new_temp_dir(**kwargs):
    return DirectoryPath(tempfile.mkdtemp(**kwargs) + '/')



from setuptools import setup

setup(
        name             = 'autopaths',
        version          = '1.3.8',
        description      = 'autopaths is a python package for dealing with file paths and automation.',
        long_description = open('README.md').read(),
        long_description_content_type = 'text/markdown',
        license          = 'MIT',
        url              = 'http://github.com/xapple/autopaths/',
        author           = 'Lucas Sinclair',
        author_email     = 'lucas.sinclair@me.com',
        packages         = ['autopaths'],
    )
